from sklearn.metrics import mean_squared_error
from numpy import sqrt
import seaborn as sns
import pandas as pd


def get_x_y_values(train, test, x_col, y_col):
    return train[x_col], train[y_col], test[x_col], test[y_col]


def get_train_test(data, test_size=0.66, viz=False, sort=False, return_array=False):
    """data: pandas.DataFrame where columns = features
    sort: whether data should be sorted (relevant as this is Ts)
    Returns train, test as pandas.DataFrame or numpy.ndarray depending on return_array
    """

    assert isinstance(data, pd.DataFrame)

    if data.shape[1] > 1:
        print('The dataframe has more than one column, sure that you have excluded the data-col?')
        print(data.head(3))

    orig_data = data.copy()

    if isinstance(test_size, float):
        train_size = 1 - test_size
        split_row = int(orig_data.shape[0]*train_size)
        train = orig_data[0:split_row]
        test = orig_data[split_row:]
    else:
        train_size = orig_data.shape[0]-test_size
        train = orig_data[0:train_size]
        test = orig_data[train_size:]

    assert (train.shape[0] + test.shape[0]) == orig_data.shape[0]

    if sort:
        # relevant as this is Ts
        print('not yet implemented')

    if viz:
        # to assert that the TS are continuos after plotting
        sns.lineplot(data=train, color='green')
        sns.lineplot(data=test, color='red')

    if return_array:
        train = train.to_numpy()
        test = test.to_numpy()

    return train, test


def get_lagged_values(data, calc_col, forecast_days, drop_na=True):
    """

    data = dataframe
    Returns a dataframe (reframed_prediction_horizon) where each column is the next value for the previous column


    don't drop NAs if I will only use test-set as test-set is not affected by the NAs (if test-set-split starts after
    train)

    Example:

    forecast_days = 7

    data:

        Date        temperature
    0   1981-01-01  20.7
    1   1981-01-02  17.9
    2   1981-01-03  18.8
    3   1981-01-04  14.6
    4   1981-01-05  15.8
    5   1981-01-06  15.8
    6   1981-01-07  15.8
    7   1981-01-08  17.4
    8   1981-01-09  21.8
    9   1981-01-10  20.0
    10  1981-01-11  16.2
    11  1981-01-12  13.3
    12  1981-01-13  16.7
    13  1981-01-14  21.5

    reframed_prediction_horizon:

    t       t+1     t+2     t+3     t+4     t+5     t+6     t+7
    NaN     20.7    17.9    18.8    14.6    15.8    15.8    15.8
    20.7    17.9    18.8    14.6    15.8    15.8    15.8    17.4
    17.9    18.8    14.6    15.8    15.8    15.8    17.4    21.8
    18.8    14.6    15.8    15.8    15.8    17.4    21.8    20.0
    14.6    15.8    15.8    15.8    17.4    21.8    20.0    16.2
    15.8    15.8    15.8    17.4    21.8    20.0    16.2    13.3
    15.8    15.8    17.4    21.8    20.0    16.2    13.3    16.7
    15.8    17.4    21.8    20.0    16.2    13.3    16.7    21.5
    17.4    21.8    20.0    16.2    13.3    16.7    21.5    25.0
    21.8    20.0    16.2    13.3    16.7    21.5    25.0    20.7
    20.0    16.2    13.3    16.7    21.5    25.0    20.7    20.6
    16.2    13.3    16.7    21.5    25.0    20.7    20.6    24.8
    13.3    16.7    21.5    25.0    20.7    20.6    24.8    17.7
    16.7    21.5    25.0    20.7    20.6    24.8    17.7    15.5
    """

    reframed_prediction_horizon = pd.DataFrame()
    prediction_base = data.copy()

    reframed_prediction_horizon[calc_col + '+0'] = prediction_base[calc_col].shift(1)
    reframed_prediction_horizon[calc_col + '+1'] = prediction_base[calc_col]

    for shift in range(1, forecast_days):
        reframed_prediction_horizon[calc_col + '+' + str(shift + 1)] = prediction_base[calc_col].shift(-shift)

    if drop_na:
        reframed_prediction_horizon = reframed_prediction_horizon.dropna().reset_index(drop=True)

    return reframed_prediction_horizon


def get_baseline(data, calc_col, viz=False, test_size=0.66, show_explanation=False, drop_na=True):
    """Uses persistence_algorithm to calc RMSE to get baseline value for forecasting performance
    data: pandas.DataFrame with calc_col
        calc_col: column that should be predicted

    # Steps in the persistence_algorithm

    1. Transformn into supervised problem (using one lag)
    2. Get test-data (inkl. x/y as this is a supervised problem now) (train will also be generated but not used for
    the predictions)
    3. Define persistence_algorithm; here: x=y, and x=y --> x=t0, y=t+1, i. e. the persistence_algorithm predicts the
    next value at any given step)
    4. Make forecast + calc rmse

    See: Introduction to Time Series Forecasting with Python How to Prepare Data and Develop Models to Predict the
    Future by Jason Brownlee
    """
    # convert into supervised problem (using one lag)

    assert isinstance(data, pd.DataFrame)
    data_orig = data.copy()

    data_new = get_lagged_values(data_orig, calc_col, 1, drop_na=drop_na)

    # get test-data
    train_data, test_data = get_train_test(data_new, test_size)

    # get x+y
    train_x, train_y, test_x, test_y = get_x_y_values(train_data, test_data, calc_col + '+0', calc_col + '+1')

    # get predictions using persistence_algorithm
    """if the train_set has two lag values, then 
    x = today
    y = tomorrow
    
    by predicting x for y (see rmse below) we assume that what happened in the previous time step will be the same
    as what will happen in the next time step
    """
    predictions = test_x

    # calc rmse
    rmse = sqrt(mean_squared_error(test_y, test_x))

    print('Test RMSE: %.3f' % rmse)

    if show_explanation:
        print("Consider: "
              "* the model should be one-step behind reality"
              "* a REAL predicted model should be stationary, however,"
              "this model won't be (provided the input data isn't either)"
              "* use Augmented Dickey-Fuller test to check for stationarity"
              )

    if viz == 'test+predict':
        sns.set(rc={'figure.figsize': (18, 6)})
        sns.lineplot(data=test_y, color='green', label='test_y')
        sns.lineplot(data=predictions, color='red', label='predictions')
    if viz == 'all':
        sns.set(rc={'figure.figsize': (18, 6)})
        sns.lineplot(data=train_y, color='blue', label='train_y')
        sns.lineplot(data=predictions, color='red', label='predictions')
        sns.lineplot(data=test_y, color='green', label='test_y')

    residuals = pd.DataFrame(test_y.to_numpy() - predictions.to_numpy())
    return train_x, train_y, test_x, test_y, predictions, rmse, residuals



from newsapi import NewsApiClient


# function to get News for NewsApi
def get_articles_from_newsapi_with_qintitle(qintitle=None
                                            , article_language=None
                                            , from_param='2020-01-01'  # yyyy-mm-dd
                                            , to_param='2020-31-12'
                                            , newsapi_key=''
                                            ):
    all_articles_all_pages_df = pd.DataFrame()
    suchseite = 1
    anzahl_artikel_gefunden_von_api = 1

    newsapi = NewsApiClient(api_key=newsapi_key)

    while anzahl_artikel_gefunden_von_api > 0:
        all_articles_current_page_respone = newsapi.get_everything(
            from_param=from_param
            , to=to_param
            , language=article_language
            , page_size=100
            , page=suchseite
            , qintitle=qintitle
        )

        all_articles_current_page = all_articles_current_page_respone['articles']
        anzahl_artikel_gefunden_von_api = len(all_articles_current_page)
        all_articles_current_page_df = pd.DataFrame.from_dict(all_articles_current_page)
        all_articles_all_pages_df = all_articles_all_pages_df.append(all_articles_current_page_df)

        suchseite += 1
        if suchseite == 100:
            break
    return all_articles_all_pages_df


def get_articles_from_newsapi_with_source(
        from_param='2020-01-01'
        , to_param='2020-31-12'
        , newsapi_key=''
        , article_source_ids=''
):
    """Downloads articles from newsapi using sources

    :param from_param: Format: yyyy-mm-dd
    :type from_param: str

    :param from_param: Format: yyyy-mm-dd
    :type from_param: str

    :param newsapi_key:
    :type from_param: str

    :param article_source_ids:
    :type from_param: str
    """

    all_articles_all_pages_df = pd.DataFrame()
    suchseite = 1
    anzahl_artikel_gefunden_von_api = 1

    newsapi = NewsApiClient(api_key=newsapi_key)

    while anzahl_artikel_gefunden_von_api > 0:
        all_articles_current_page_respone = newsapi.get_everything(
            from_param=from_param
            , to=to_param
            , page_size=100
            , page=suchseite
            , sources=article_source_ids
        )

        all_articles_current_page = all_articles_current_page_respone['articles']
        anzahl_artikel_gefunden_von_api = len(all_articles_current_page)
        all_articles_current_page_df = pd.DataFrame.from_dict(all_articles_current_page)
        all_articles_all_pages_df = all_articles_all_pages_df.append(all_articles_current_page_df)

        suchseite += 1
        if suchseite == 100:
            break
    return all_articles_all_pages_df



import nltk
import re
import string
import codecs
from collections import Counter


def worthaeufigkeiten_berechnen(text):
    # text = String
    # haeufigkeiten berechnen
    haeufigkeiten_dict = dict(Counter(text.split()))

    # dict sortieren
    haeufigkeiten_list_sorted = sorted(haeufigkeiten_dict.items(), key=lambda kv: kv[1], reverse=True)

    return haeufigkeiten_dict, haeufigkeiten_list_sorted


def remove_stopwords(text
                     , custom_stopwords_file='industry_classifier_custom_stopword_list'
                     , default_stopwords_language='german'):
    """

    :param default_stopwords_language: define the language based on which the stopwords will be removed
    :param text: from which text the stopwords should be removed
    :param custom_stopwords_file: string, optional
        Voller Pfad zu einer benutzerdefinierten Datei mit Stoppwörtern.
        Dateiformat muss .csv sein.
        Inhalt der .csv-Datei: jede Zeile ist eine Stoppwort.

        Beispiel für Inhalt einer .csv-Datei:

        070316umschlagfu
        100prozentzielwert
        12f
        132r1


    :return: Cleaned text as string with one space between each token/word
    """
    default_stopwords = set(nltk.corpus.stopwords.words(default_stopwords_language))
    all_stopwords = default_stopwords

    if custom_stopwords_file:
        custom_stopwords = set(codecs.open(custom_stopwords_file, 'r', 'utf-8').read().splitlines())
        all_stopwords = default_stopwords | custom_stopwords
    else:
        pass
    tokens = tokenize_text(text)
    filtered_tokens = [token for token in tokens if token not in all_stopwords and len(token) > 1]
    filtered_text = ' '.join(filtered_tokens)
    return filtered_text


def tokenize_text(text):
    """
    Returns text tokenized as words
    :param str text: Text to be tokenized
    :return: Words
    :rtype: String


    Examples
    --------
    text = ' drehmomentwandlervorrichtung, insbesondere für einen antriebsstrang eines kraftfahrzeugs  '

    print(tokenize_text(titel[0]))
    ['drehmomentwandlervorrichtung',
    ',',
    'insbesondere',
    'für',
    'einen',
    'antriebsstrang',
    'eines',
    'kraftfahrzeugs']

    """
    tokens = nltk.word_tokenize(text)
    tokens = [token.strip() for token in tokens]
    return tokens


def replace_special_characters(text, substitutionsstring=' '):
    """Entfernt special_characters gänzlich.

    special_characters = '!"#$%&\'()*+,-./:;<=>?@[\\]^_`{|}~' - siehe <pattern>
    Beispiele:
    Vorher: http //www.linkedin.com/company/daimler/
    Nachher: wwwlinkedincomcompanydaimler
    """
    tokens = tokenize_text(text)
    pattern = re.compile('[{}]'.format(re.escape(string.punctuation)))
    filtered_tokens = filter(None, [pattern.sub(substitutionsstring, token) for token in tokens])
    filtered_text = ' '.join(filtered_tokens)
    return filtered_text


def stem_words(text, spacy_model):
    spacy_doc = spacy_model(text)
    stemmed_tokens = [token.lemma_ for token in spacy_doc]
    stemmed_text = ' '.join(stemmed_tokens)
    return stemmed_text


def keep_text_characters(text):
    filtered_tokens = []
    tokens = tokenize_text(text)
    for token in tokens:
        if re.search('[a-zA-Z]', token):
            filtered_tokens.append(token)
    filtered_text = ' '.join(filtered_tokens)
    return filtered_text


def remove_digits_from_text(text):
    # make this '49die' into this 'die'
    pattern = '[0-9]'
    tokens_original = tokenize_text(text)
    tokens = [re.sub(pattern, '', token) for token in tokens_original]
    text = ' '.join(tokens)
    return text


def remove_links(text):
    text = re.sub(r'http\S+', '', text)
    return text


def normalize_text(text
                    , tokenize=True
                    , only_text_chars=True
                    , filter_stopwords=True
                    , custom_stopwords_file_name=''
                    , word_stemming=True
                    , rm_special_char=True
                    , rm_links=True
                    , rm_digits_from_text=True
                    , spacy_model_name='de_core_news_sm'
                    , default_stopwords_language='german'
                   , substitution_string = ''
                   ):
    """

    :param default_stopwords_language: define the language based on which the text will be normalized
    :param spacy_model_name: which spacy model should be used if necessary
    :type rm_links: bool whether links should be fully removed
    :param rm_digits_from_text: bool whether digits should be fully removed
    :type rm_special_char: bool
    :param text: the text that should be normalized
    :param tokenize: if the text should be returned as tokens or joined string
    :param only_text_chars:
    :param filter_stopwords:
    :param custom_stopwords_file_name:
    :param word_stemming:
    :param substitution_string: used in replace_special_characters to define with what to replace the spec. char.
    :return: normalized text as token or string depending on tokenize
    """
    # Lowercase all words
    text = text.lower()
    if word_stemming:
        text = stem_words(text, spacy_model_name)
    if rm_special_char:
        text = replace_special_characters(text, substitution_string)
    if only_text_chars:
        text = keep_text_characters(text)
    if filter_stopwords:
        text = remove_stopwords(text, custom_stopwords_file_name, default_stopwords_language)
    else:
        pass
    if rm_digits_from_text:
        text = remove_digits_from_text(text)
    if rm_links:
        text = remove_links(text)
    if tokenize:
        text = tokenize_text(text)

    return text



import pandas as pd


def get_value_counts_as_df(data_frame, column_name_grouping, column_name_count='count'):
    """Return a data frame containing counts of unique values
    :param data_frame: Pandas data frame to process
    :param column_name_grouping: the column for which the unique values are counted
    :param column_name_count: the name of the column that contains the counts for column_grouping
    :return: grouped_values: data frame with column_grouping and column_count as headers

    Example:

    column_name_grouping = 'verified_purchase'
    column_name_count = 'count'
    grouped_values =
        verified_purchase	count
    0	Y	                371539
    1	N	                256810
    """
    grouped_values = data_frame[column_name_grouping].value_counts().rename_axis(column_name_grouping).reset_index(
        name=column_name_count)
    grouped_values.sort_values(by=[column_name_count], inplace=True, ascending=False)
    grouped_values.reset_index(drop=True, inplace=True)

    return grouped_values



from azure.ai.textanalytics import TextAnalyticsClient
from azure.core.credentials import AzureKeyCredential
import logging
import pandas as pd
import ast


def get_text_analytics_client(azure_key, azure_endpoint):
    ta_credential = AzureKeyCredential(azure_key)
    text_analytics_client = TextAnalyticsClient(
            endpoint=azure_endpoint,
            credential=ta_credential)
    return text_analytics_client


def get_entities(documents, client):
    """
    :param documents: should look like this: text = ['This sentence talks about iPhones and Sushi.']
    :param client:
    :return:
    """
    try:
        result = client.recognize_entities(documents=documents)[0]
        entities = result.entities
        return entities
    except Exception as err:
        print("Encountered exception. {}".format(err))


def get_entities_for_category(entities_base, entity_category):
    """
    :param entity_category: could be for instance 'Organization'
    :param entities_base: entities as returned from azure, has all entities, from which entity_category will be filtered
    :return: entities_filtered

    Example:

    entity_category = 'Product'
    entities_base = [CategorizedEntity(text=Buches, category=Product, subcategory=None, confidence_score=0.86),
                     CategorizedEntity(text=Mädelsabend, category=Event, subcategory=None, confidence_score=0.81)]
    entities_filtered = '['Buches']'
    """
    entities_filtered = ''
    try:
        entities_filtered = [entity.text for entity in entities_base if entity.category == entity_category]
    except:
        pass
    return entities_filtered


def get_sentiment(client, documents, show_documents_stats=False, documents_language='de'):
    """
    :param client: TextAnalyticsClient
    :param documents: for instance ["this is the text", "this is another doc"]
    :param show_documents_stats: see official doc
    :param documents_language: see official doc
    :return: sentiment_overall_all_docs, sentiment_sentences_all_docs
    pandas data frames for the overall doc and each sentence in each doc

    """

    def _get_individual_sentiments(data_frame, tgt_column_prefix, src_column):
        individual_sentiments = [
            'positive'
            , 'neutral'
            , 'negative'
        ]

        for suffix in individual_sentiments:
            data_frame[tgt_column_prefix + '_' + suffix] = data_frame.apply(
                lambda x: ast.literal_eval(x[src_column]).get(suffix), axis=1)

        return data_frame

    logger = logging.getLogger("azure.core.pipeline.policies.http_logging_policy")
    logger.setLevel(logging.WARNING)
    response_set = client.analyze_sentiment(
        documents=documents
        , show_stats=show_documents_stats
        , language=documents_language
    )

    sentiment_sentences_all_docs = pd.DataFrame()
    sentiment_overall_all_docs = pd.DataFrame()
    overall_response_keys = [
        'sentiment'
        , 'confidence_scores'
    ]
    sentence_response_keys = [
        'text'
        , 'sentiment'
        , 'confidence_scores'
    ]

    def _extract_from_response(df, df_index, response_keys, column_prefix, response_dict):
        for response_key in response_keys:
            df.loc[df_index, column_prefix + '__' + response_key] = str(
                response_dict.get(response_key))
        return df

    for doc_idx, response in enumerate(response_set):
        df_idx = sentiment_overall_all_docs.shape[0]
        overall__id = response.get('id')

        sentiment_overall_all_docs = _extract_from_response(
            sentiment_overall_all_docs, df_idx, overall_response_keys, 'overall', response
        )
        sentiment_overall_all_docs.loc[df_idx, 'doc_text'] = documents[doc_idx]
        sentiment_overall_all_docs.loc[df_idx, 'overall__id'] = overall__id

        # do stuff on sentence level
        sentences = response.get('sentences')
        if not sentences:
            continue

        for sentence in sentences:
            df_idx = sentiment_sentences_all_docs.shape[0]
            sentiment_sentences_all_docs = _extract_from_response(
                sentiment_sentences_all_docs, df_idx, sentence_response_keys, 'sentence', sentence
            )
        sentiment_sentences_all_docs['overall__id'] = overall__id

    sentiment_overall_all_docs = _get_individual_sentiments(
        sentiment_overall_all_docs
        , 'overall__confidence_score'
        , 'overall__confidence_scores'
    )
    sentiment_sentences_all_docs = _get_individual_sentiments(
        sentiment_sentences_all_docs
        , 'sentence__confidence_score'
        , 'sentence__confidence_scores'
    )

    return sentiment_overall_all_docs, sentiment_sentences_all_docs


def extract_from_response_to_df(
        original_df
        , response_keys
        , column_prefix
        , response
        , original_document
):
    """Extracts values from the response-dict and writes into a pandas data_frame

    :param original_df: the original pandas data_frame into which to write the new key phrases from response_dict
    :param response_keys: for which keys to extract the value from the response-dict
    :param column_prefix: how the data_frame column that contains the value for response_keys should be named
    :param response: the response-dict from which to extract stuff
    :param original_document: contains the document from which the key phrases were extracted
    :return: df where each column is one response_key + the original document
    """
    key_phrases = pd.DataFrame(response.get('key_phrases'), columns=[column_prefix + '__key_phrases'])

    """iterate over each key, get the value for this key from the dict and write this value into the data_frame
    str() as some values are not string and thus can't be inserted into the df
    """
    for response_key in response_keys:
        key_phrases[column_prefix + '__' + response_key] = str(response.get(response_key))

    # set the original doc to the data_frame
    key_phrases[column_prefix + '__' + 'original_document'] = original_document

    # add the key_phrases to the original df
    original_df = original_df.append(key_phrases)
    original_df.reset_index(drop=True, inplace=True)
    return original_df


def get_key_phrases(client, documents, show_documents_stats=False, documents_language='de'):
    logger = logging.getLogger("azure.core.pipeline.policies.http_logging_policy")
    logger.setLevel(logging.WARNING)

    try:
        response_set = client.extract_key_phrases(
            documents=documents
            , show_stats=show_documents_stats
            , language=documents_language
        )

        key_phrases_documents = pd.DataFrame()

        # each response is for one doc
        for response_idx, response in enumerate(response_set):
            if response.is_error:
                print(response.id, response.error)
                continue

            response_keys = response.keys()
            response_keys.remove('key_phrases')

            key_phrases_documents = extract_from_response_to_df(
                key_phrases_documents
                , response_keys
                , 'keywords_response'
                , response
                , documents[response_idx]
            )
        return key_phrases_documents

    except Exception as err:
        print("Encountered exception. {}".format(err))